#include "c3dtypes.h"
#include "raylib.h"
#include "raymath.h"
#include "reader.h"
#include "float.h"
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "arrayfuncs.h"
#include "vecfunc.h"

const int RENDERWIDTH = 1920;
const int RENDERHEIGHT = 1080;

const int HALFWIDTH = RENDERWIDTH / 2;
const int HALFHEIGHT = RENDERHEIGHT / 2;

const int SCREENWIDTH = 1280;
const int SCREENHEIGHT = 720;

const float WIDTHSCALE = (float)RENDERWIDTH / SCREENWIDTH;
const float HEIGHTSCALE = (float)RENDERHEIGHT / SCREENHEIGHT;

const float fov = 70.0 / 180.0 * 3.14159265369;
const float half_fov = fov / 2;

float proj; // HALFWIDTH / tan(half_fov) Inited later

bool printdebug;

void PrintTri2D(Tri2D t) {
    printf("{(TRI2D) (Color: %d %d %d) A: (%f, %f, d:%f), B: (%f, %f, d:%f), C:(%f,%f, d:%f) }\n ",
           t.color.r, t.color.g, t.color.b, t.a.x, t.a.y, t.adepth, t.b.x, t.b.y, t.bdepth, t.c.x,
           t.c.y, t.cdepth);
}

Vector2 Conv2DCenteredToScreen(Vector2 v) {
    Vector2 returnvector;
    returnvector.x = v.x + HALFWIDTH;
    returnvector.y = v.y + HALFHEIGHT;
    return returnvector;
}


static inline Vector2 ConvScreenToCentered(Vector2 v) {
    Vector2 returnvector;
    returnvector.x = v.x - HALFWIDTH;
    returnvector.y = v.y - HALFHEIGHT;
    return returnvector;
}


float Sign(Vector2 *v1, Vector2 *v2, Vector2 *v3) {
    return (v1->x - v3->x) * (v2->y - v3->y) - (v2->x - v3->x) * (v1->y - v3->y);
}

static inline bool IsInTri(Tri2D tri, Vector2 p) {
    float d1, d2, d3;
    bool has_neg, has_pos;

    d1 = Sign(&p, &tri.a, &tri.b);
    d2 = Sign(&p, &tri.b, &tri.c);
    d3 = Sign(&p, &tri.c, &tri.a);

    has_neg = (d1 < 0) || (d2 < 0) || (d3 < 0);
    has_pos = (d1 > 0) || (d2 > 0) || (d3 > 0);

    return !(has_neg && has_pos);
}
static inline Vector3 Tri2DBaryAtPoint(Tri2D *t, Vector2 p) {
    Vector3 retvec;
    Vector2 v0 = Vector2Subtract(t->b, t->a); // cachable
    Vector2 v1 = Vector2Subtract(t->c, t->a); // cacheable
    Vector2 v2 = Vector2Subtract(p, t->a);
    float d00 = Vector2DotProduct(v0, v0); // cacheable
    float d01 = Vector2DotProduct(v0, v1); // cacheable
    float d11 = Vector2DotProduct(v1, v1); // cahceable
    float d20 = Vector2DotProduct(v2, v0);
    float d21 = Vector2DotProduct(v2, v1);
    float denom = d00 * d11 - d01 * d01; // cache
    retvec.y = (d11 * d20 - d01 * d21) / denom; //fixed ordering, 
    retvec.z = (d00 * d21 - d01 * d20) / denom;
    retvec.x = 1.0 - retvec.y - retvec.z;
    return retvec;
}

Vector2 BaryAndTritoPoint(Tri2D *t, Vector3 bary) {
    Vector2 retvec;
    retvec.x = bary.x * t->a.x + bary.y * t->b.x + bary.z * t->c.x;
    retvec.y = bary.y * t->a.y + bary.y * t->b.y + bary.z * t->c.y;
    return retvec;
}

void Vector2Integate(Vector2 *v) {
    v->x = (int)v->x;
    v->y = (int)v->y;
}

static inline float InvDepthAtBary(Tri2D *t, Vector3 bary) {
    float firstrepz = 1.0 / t->adepth;
    float secondrepz = 1.0 / t->bdepth;
    float thirdrepz = 1.0 / t->cdepth;

    float finrepz = (firstrepz * bary.x) + (secondrepz * bary.y) + (thirdrepz * bary.z);

    return  finrepz;
}

double Min(double a, double b) {
    if (a > b) {
        return b;
    }
    return a;
}

double Max(double a, double b) {
    if (a < b) {
        return b;
    }
    return a;
}

static inline int IndexOfZBuff(int row, int col) {
    // asfd
    return row + RENDERWIDTH * col;
}

// sort triangle verts so that point A is the "highest" point (lowest y val) and
// point C is the "lowest" pont (highest y val)
void Tri2DSortByY(Tri2D *t) {
    Vector2 temp;
    Vector2 tempuv;
    float tempdepth;
    if (t->a.y > t->b.y) {
        temp = t->a;
        tempuv = t->auv;
        tempdepth = t->adepth;

        t->a = t->b;
        t->auv = t->buv;
        t->adepth = t->bdepth;

        t->b = temp;
        t->buv = tempuv;
        t->bdepth = tempdepth;
    }
    if (t->b.y > t->c.y) {
        temp = t->b;
        tempuv = t->buv;
        tempdepth = t->bdepth;

        t->b = t->c;
        t->buv = t->cuv;
        t->bdepth = t->cdepth;

        t->c = temp;
        t->cuv = tempuv;
        t->cdepth = tempdepth;
    }
    if (t->a.y > t->b.y) {
        temp = t->a;
        tempuv = t->auv;
        tempdepth = t->adepth;

        t->a = t->b;
        t->auv = t->buv;
        t->adepth = t->bdepth;

        t->b = temp;
        t->buv = tempuv;
        t->bdepth = tempdepth;
    }
}

void V2Print(Vector2 v) {
    // ra
    printf("%f, %f\n", v.x, v.y);
}

Vector2 GetUVFromTriAndBary(Vector3 bary, Tri2D * tri, float z){
	Vector2 ta = Vector2Scale(tri->auv,   (1.0 / tri->adepth ));
	Vector2 tb = Vector2Scale(tri->buv,   (1.0 / tri->bdepth ));
	Vector2 tc = Vector2Scale(tri->cuv,   (1.0 / tri->cdepth ));

	Vector2 AB = Vector2Add( Vector2Scale(ta, bary.x), Vector2Scale(tb, bary.y))  ;
	Vector2 ABC = Vector2Add(AB, Vector2Scale(tc, bary.z));
	return Vector2Scale(ABC, z);
}

void DrawScanline(Zee *zee, Tri2D *trianglepointer, double start, double end, int scanline) {
    float depth = 0;
    Vector2 UVpos;
    UVpos.x = 0;
    UVpos.y = 0;
    Tri2D UVRangeTri;
    UVRangeTri.a = trianglepointer->auv;
    UVRangeTri.b = trianglepointer->buv;
    UVRangeTri.c = trianglepointer->cuv;
    /* V2Print(UVRangeTri.a); */
    /* V2Print(UVRangeTri.b); */
    /* V2Print(UVRangeTri.c); */
    
    Vector3 f1baryatpoint = Tri2DBaryAtPoint(trianglepointer, (Vector2){Max(start, 0), scanline});
    Vector3 f1baryatpoint2 =
        Tri2DBaryAtPoint(trianglepointer, (Vector2){Max(start, 0) + 1, scanline});

    double f1invdepth = InvDepthAtBary(trianglepointer, f1baryatpoint);
    double f1invdepth2 =  InvDepthAtBary(trianglepointer, f1baryatpoint2);


	Vector2 tex1point = GetUVFromTriAndBary(f1baryatpoint, trianglepointer, 1.0 / f1invdepth);
    Vector2 tex2point = GetUVFromTriAndBary(f1baryatpoint2, trianglepointer, 1.0 / f1invdepth2);

    double dinvslope = f1invdepth2 - f1invdepth; // change in store_depth per pixel

    Vector2 UVSlope = Vector2Subtract(tex2point, tex1point);
    //	printf("UVSLOPE: X%f: Y:%f\n", UVSlope.x, UVSlope.y);
    for (int i = (Max(start, 0)); i < (Min(end, RENDERWIDTH)); i++) {

        // UNCOMMENT TO DO TRUE NON SLOPE BASED DEPTH CALC:
        Vector3 baryatpoint = Tri2DBaryAtPoint(trianglepointer,  (  (Vector2){i, scanline}));
        //depth = 1.0 / (f1invdepth + (dinvslope * ((float)i - (float)(Max(start, 0)))));
        depth = 1.0 / InvDepthAtBary(trianglepointer, baryatpoint);

		
		//UNCOMMENT TO DO REAL UV CALC
		UVpos = BaryAndTritoPoint(&UVRangeTri, baryatpoint);
		UVpos = GetUVFromTriAndBary(baryatpoint, trianglepointer, depth);
		
		//V2Print(UVpos);	
		
        //UVpos = Vector2Add(tex1point, Vector2Scale(UVSlope, ((float)i - (float)(Max(start, 0)))));
		Vector2 MAX_VEC2 = (Vector2) {FLT_MAX, FLT_MAX};
		UVpos = Vector2Clamp(UVpos, Vector2Zero(), MAX_VEC2);
        if (depth < zee[IndexOfZBuff(i, scanline)].depth) {

            CMaterial *mat = trianglepointer->material;
			
            if (mat != NULL) {

                int y = (int)(UVpos.y * (float)mat->h) % mat->h; // keep within range because of bug
                int x = (int)(UVpos.x * (float)mat->w) % mat->w;
                // printf("plotting point x: %f, y: %d\n", UVpos.x,y);
                // int pitch =
                zee[IndexOfZBuff(i, scanline)].c = trianglepointer->material->tex[(y * mat->w) + x];
            } else {
                zee[IndexOfZBuff(i, scanline)].c = trianglepointer->color;
            }
            zee[IndexOfZBuff(i, scanline)].depth = depth;
            //zee[IndexOfZBuff(i, scanline)].tri = *trianglepointer;
        }
    }
}

// Draws triangle with a flat top. Note A and B must be the top points with C
// being the bottom "spike"
void FillTopFlatZbuffer(Zee *zee, Tri2D *t, Tri2D *tp) {
    if (t->b.x < t->a.x) {
        Vector2 e = t->b;
        t->b = t->a;
        t->a = e;
    }
    // Becasue we are trying to get the x values in terms of the y values, we
    // need inverse slope
    double atocslopeinv =
        (t->c.x - t->a.x) / (t->c.y - t->a.y); // dif in x from start to end with a and c
    double btocslopinv =
        (t->c.x - t->b.x) / (t->c.y - t->b.y); // dif in x from start to end with b and c

    // start at bottom of triangle (point c) so that we do not need to determine
    // which vertex is on which side and increment it with its proper slope
    double curx1 = t->c.x;
    double curx2 = t->c.x;

    int scanbottom = (t->c.y);
    int scantop = (Max(t->a.y, 0));

    for (int scanline = scanbottom; scanline >= scantop; scanline--) {
        if (0 <= scanline && scanline < RENDERHEIGHT) {
            DrawScanline(zee, tp, curx1, curx2, scanline);
        }
        curx1 -= atocslopeinv; // subtract because we are working backwards (reason
                               // why we start with point c in slope equtn)
        curx2 -= btocslopinv;
    }
}

// Draws triangle with a flat bottomp. Note B and C must be the bottom points
// with A being the top "spike"
void FillBottomFlatZbuffer(Zee *zee, Tri2D *t, Tri2D *tp) {
    if (t->c.x < t->b.x) {
        Vector2 e = t->c;
        t->c = t->b;
        t->b = e;
    }
    // Becasue we are trying to get the x values in terms of the y values, we
    // need inverse slope
    double atobslopeinv =
        (t->b.x - t->a.x) / (t->b.y - t->a.y); // dif in x from start to end with a and c
    double atocslopeinv =
        (t->c.x - t->a.x) / (t->c.y - t->a.y); // dif in x from start to end with b and c

    // start at top of triangle (point c) so that we do not need to determine
    // which vertex is on which side and increment it with its proper slope
    double curx1 = t->a.x;
    double curx2 = t->a.x;

    int scantop = (t->a.y); // start must always be the true beging of the triangle, otherwise
                            // everything will be offset
    int scanbottom =
        (Min(t->c.y, RENDERHEIGHT)); // we can stop rendering as soon as it goes off screen

    for (int scanline = scantop; scanline <= scanbottom;
         scanline++) { // TODO: Possibly more optimization possible here, use linear correspondance
                       // for y, not just x to get depth
        if (0 <= scanline && scanline < RENDERHEIGHT) {
            Tri2D nt = *tp;
            nt.color = RED;

            DrawScanline(zee, tp, curx1, curx2, scanline);
        }
        curx1 += atobslopeinv;
        curx2 += atocslopeinv;
    }
}

/* void DrawTriZuff2simple(Zee *zbuf, Tri2D *tex){ */
/* 	for (int y = 0; y < RENDERHEIGHT; y++){ */
/* 		for (int x = 0; x < RENDERWIDTH; x++){ */
/* 			if (IsInTri(*tex, (Vector2){x,y})) { */
/* 				Vector2 p = (Vector2) {x,y}; */
/* 				Vector3 baryatpoint = Tri2DBaryAtPoint(tex,  (  (Vector2){x, y})); */
/* 				Vector2 bp = BaryAndTritoPoint(tex, baryatpoint); */
/* 				//if (bp.x != p.x || bp.y != p.y) { */
/* 			//		printf("Gotcha:\n"); */
/* 			//		Vector2Print(p); */
/* 			//		Vector2Print(bp); */
/* 			//		printf("GotchaEND:\n"); */

/* //				}	 */
/* 				float depth = InvDepthAtBary(tex, baryatpoint); */

/* 				if (depth < zbuf[IndexOfZBuff(x, y)].depth) { */
/* 					zbuf[IndexOfZBuff(x, y)].depth = depth; */
/* 					zbuf[IndexOfZBuff(x, y)].c = tex->color; */

/* 				} */
/* 				} */
/* 		} */
/* 	} */
/* } */




void AnotateTri(Tri2D *tex) {
	char linea[500];
	char lineb[500];
	char linec[500];
	sprintf(linea, "UV: %f, %f;", tex->auv.x, tex->auv.y);
	sprintf(lineb, "UV: %f, %f;", tex->buv.x, tex->buv.y);
	sprintf(linec, "UV: %f, %f;", tex->cuv.x, tex->cuv.y);

	DrawText(linea, tex->a.x, tex->a.y, 60, WHITE);
	DrawText(lineb, tex->b.x, tex->b.y, 60, WHITE);
	DrawText(linec, tex->c.x, tex->c.y, 60, WHITE);
	
}

//the issues is the tt pointer changes, so the thing does not reference properly for debugging
void DrawTriZuff(Zee *zbuf, Tri2D *tex) {
    Tri2D tt = *tex;
    Tri2D *t = &tt;

    Tri2DSortByY(t);

	Tri2D cpy = *t;
	
    Vector2Integate(&t->a);
    Vector2Integate(&t->b);
    Vector2Integate(&t->c);

	//AnotateTri(tex);

    CMaterial *mat = tex->material;
    /* int y = (int) UVpos.y % mat->h; // keep within range because of bug */
    /* int x = (int) UVpos.x % mat->w; */
    // printf("%s\n", mat->name);
    if ((t->a.y > RENDERHEIGHT && t->c.y > RENDERHEIGHT) || (t->a.y < 0 && t->c.y < 0)) {
        return;
    }

    if (t->b.y == t->c.y) { // if bottom of triangle is flat
        FillBottomFlatZbuffer(zbuf, t, tex);
    } else if (t->a.y == t->b.y) { // if top of triangle is flat

        FillTopFlatZbuffer(zbuf, t, tex);

    } else {        // funny split tri
        Vector2 v4; // v4 is the vertex on the line between a and c. It is used
                    // to split the triangle into a top and bottom
        v4.y = t->b.y;
        float slope =
            (float)((t->c.x) - (t->a.x)) / ((float)(t->c.y - t->a.y)); // get slope in run over rise
                                                                       // becasue we need to find x
        float changeiny = (float)(t->b.y - t->a.y);
        float officalxpos = t->a.x + (slope * changeiny);
        v4.x = officalxpos;
        Tri2D bottomflattrires;
        bottomflattrires.a = t->a;
        bottomflattrires.b = t->b;
        bottomflattrires.c = v4;

        bottomflattrires.color = t->color;
        Tri2D topflattrires;
        topflattrires.a = t->b;
        topflattrires.b = v4;
        topflattrires.c = t->c;

        topflattrires.color = t->color;

        FillBottomFlatZbuffer(zbuf, &bottomflattrires, tex);
        FillTopFlatZbuffer(zbuf, &topflattrires, tex);
    }
}

void CtrlLocalCam(LocalCam *cam, float time) {
    cam->velocity.x = 0;
    cam->velocity.y = 0;
    cam->velocity.z = 0;
    cam->angleVelocity.x = 0;
    cam->angleVelocity.y = 0;
    cam->angleVelocity.z = 0;

    if (IsKeyDown(KEY_W)) {
        Vector3 forceForward = (Vector3){0, 0, -500};
        Vector3 rotatedforce = RotateAboutY(forceForward, cam->angles.y);
        // printf("%f", cam->angles.y);
        // Vector3Print(rotatedforce);
        cam->velocity = Vector3Add(cam->velocity, rotatedforce);

        // cam->velocity.z = cam->velocity.z + 50000*time*sin(cam->angles.y);
        // cam->velocity.x = cam->velocity.x + 500*time*cos(cam->angles.y);
    }
    if (IsKeyDown(KEY_S)) {
        Vector3 forceForward = (Vector3){0, 0, 500};
        Vector3 rotatedforce = RotateAboutY(forceForward, cam->angles.y);
        cam->velocity = Vector3Add(cam->velocity, rotatedforce);

        // cam->velocity.z = cam->velocity.z + 50000*time*sin(cam->angles.y);
        // cam->velocity.x = cam->velocity.x + 500*time*cos(cam->angles.y);
    }

    if (IsKeyDown(KEY_A)) {
        Vector3 forceForward = (Vector3){-500, 0, 0};
        Vector3 rotatedforce = RotateAboutY(forceForward, cam->angles.y);
        cam->velocity = Vector3Add(cam->velocity, rotatedforce);

        // cam->velocity.z = cam->velocity.z + 50000*time*sin(cam->angles.y);
        // cam->velocity.x = cam->velocity.x + 500*time*cos(cam->angles.y);
    }
    if (IsKeyDown(KEY_D)) {
        Vector3 forceForward = (Vector3){500, 0, 0};
        Vector3 rotatedforce = RotateAboutY(forceForward, cam->angles.y);
        cam->velocity = Vector3Add(cam->velocity, rotatedforce);

        // cam->velocity.z = cam->velocity.z + 50000*time*sin(cam->angles.y);
        // cam->velocity.x = cam->velocity.x + 500*time*cos(cam->angles.y);
    }

    if (IsKeyDown(KEY_SPACE)) {
        cam->velocity.y = cam->velocity.y - 50;
    }

    if (IsKeyDown(KEY_LEFT_SHIFT)) {
        cam->velocity.y = cam->velocity.y + 50;
    }

    if (IsKeyDown(KEY_Q)) {
        cam->angleVelocity.y = cam->angleVelocity.y + time * 30;
    }
    if (IsKeyDown(KEY_E)) {
        cam->angleVelocity.y = cam->angleVelocity.y - time * 30;
    }
    if (IsKeyDown(KEY_R)) {
        cam->angleVelocity.x = cam->angleVelocity.x + time * 30;
    }
    if (IsKeyDown(KEY_F)) {
        cam->angleVelocity.x = cam->angleVelocity.x - time * 30;
    }

    if (IsKeyDown(KEY_G)) {
        printf("PosX: %f PosY: %f PosZ: %f\n", cam->position.x, cam->position.y, cam->position.z);
        printf("RotX; %f RotY: %f RotZ: %f\n", cam->angles.x, cam->angles.y, cam->angles.z);
    }
}

void LocalCamApplyVelo(LocalCam *cam, float time) {
    cam->position.x = cam->position.x + cam->velocity.x * time;
    cam->position.y = cam->position.y + cam->velocity.y * time;
    cam->position.z = cam->position.z + cam->velocity.z * time;

    cam->angles.x = cam->angles.x + cam->angleVelocity.x * time;
    cam->angles.y = cam->angles.y + cam->angleVelocity.y * time;
    cam->angles.z = cam->angles.z + cam->angleVelocity.z * time;
    // printf("%f %f %f\n",cam->angles.x, cam->angles.y, cam->angles.z);
}

Vector2 Conv3Dto2D(Vector3 v) {
    Vector2 returnvector;
    returnvector.x = proj * v.x / (-v.z);
    returnvector.y = proj * v.y / (-v.z);
    return returnvector;
}


static inline Vector3 TransformWithCam(Vector3 v, LocalCam *cam) {
    Vector3 returnvector;
    returnvector.x = v.x - cam->position.x;
    returnvector.y = v.y - cam->position.y;
    returnvector.z = v.z - cam->position.z;

    returnvector = RotateAboutZ(returnvector, -cam->angles.z);
    returnvector = RotateAboutY(returnvector, -cam->angles.y);
    returnvector = RotateAboutX(returnvector, -cam->angles.x);

    // printf("Before: %f %f %f, After: %f %f %f\n", v.x, v.y, v.z,
    // returnvector.x, returnvector.y, returnvector.z);
    return returnvector;
}

Tri TriTransformWithCam(Tri *t, LocalCam *cam) {
    Tri rettri = *t;
    rettri.a = TransformWithCam(t->a, cam);
    rettri.b = TransformWithCam(t->b, cam);
    rettri.c = TransformWithCam(t->c, cam);

    rettri.color = t->color;

    return rettri;
}

Tri2D ConvertTriToTri2D(Tri *t) {
    Tri2D rettri2d;
    rettri2d.a = Conv3Dto2D(t->a);
    rettri2d.b = Conv3Dto2D(t->b);
    rettri2d.c = Conv3Dto2D(t->c);

	
    rettri2d.a = Conv2DCenteredToScreen(rettri2d.a);
    rettri2d.b = Conv2DCenteredToScreen(rettri2d.b);
    rettri2d.c = Conv2DCenteredToScreen(rettri2d.c);

    rettri2d.adepth = -t->a.z;
    rettri2d.bdepth = -t->b.z;
    rettri2d.cdepth = -t->c.z;

    rettri2d.color = t->color;
    rettri2d.material = t->material;

    rettri2d.auv = t->auv;
    rettri2d.buv = t->buv;
    rettri2d.cuv = t->cuv;
    return rettri2d;
}

int main() {
    printf("%d\n", TestFunc(5));
    proj = HALFWIDTH / tan(half_fov);

    SetConfigFlags(FLAG_WINDOW_UNDECORATED);
    InitWindow(SCREENWIDTH, SCREENHEIGHT, "raylib [core] example - basic window");

    // SetWindowPosition(0,1080);
    Vector2 a = GetMonitorPosition(0);
    int mh = GetMonitorHeight(0);
    int mw = GetMonitorWidth(0);
    int w = SCREENWIDTH;
    int h = SCREENHEIGHT;
    printf("mh:%d mw:%d w:%d h:%d\n", mh, mw, w, h);
    SetWindowPosition(a.x + (0.5 * mw) - (w / 2), a.y + (0.5 * mh) - (0.5 * h));

    RenderTexture2D uiraylibtexture = LoadRenderTexture(RENDERWIDTH, RENDERHEIGHT);

    RenderTexture2D render3dtexture = LoadRenderTexture(RENDERWIDTH, RENDERHEIGHT);

    Texture2D directaccesstex;

    // Init cube model
    // TODO: Load from obj

    LocalCam camera;
    camera.position = (Vector3){0, 0, 0};
    camera.acceleration = (Vector3){0, 0, 0};
    camera.angleAcceleration = (Vector3){0, 0, 0};
    camera.angles = (Vector3){0, 0, 0};
    camera.velocity = (Vector3){0, 0, 0};
    camera.angleVelocity = (Vector3){0, 0, 0};

    Vector3 point = (Vector3){0, 0, -10};

    static Tri internaltriarray[50000];
    TriArray tarr;
    tarr.arr = internaltriarray;
    tarr.length = 0;
    TriArrayAppend(&tarr, (Tri){(Vector3){0, 0, -1000}, (Vector3){0, 800, -1000},
                                (Vector3){800, 800, -1000}, WHITE});

    TriArrayAppend(&tarr, (Tri){(Vector3){0, 0, -1010}, (Vector3){0, 800, -1010},
                                (Vector3){800, 800, -1010}, BLUE});

    static Tri internaltransformedtriarray[50000];
    TriArray TransformedTris;
    TransformedTris.arr = internaltransformedtriarray;
    TransformedTris.length = 0;

    static Tri2D internaltri2darray[50000];
    Tri2DArray Tri2Darr;
    Tri2Darr.length = 0;
    Tri2Darr.arr = internaltri2darray;

    // static Zee ZBuff[1920][1080] = {{(Zee){10000,NULL}}}; //FIXME: Stupid
    // static makes the file 32 Megs because pog
    Zee *ZBuff = malloc(RENDERHEIGHT * RENDERWIDTH * sizeof(Zee));
    static Color display[1920 * 1080 * 4];
    memset(display, 0, sizeof(display));

    Tri2D funners =
        (Tri2D){(Vector2){0, 0}, (Vector2){500, 0}, (Vector2){500, 1}, -1000, -1000, -1000, GREEN};
    Tri2D funners2 =
        (Tri2D){(Vector2){600, 0}, (Vector2){600, 500}, (Vector2){1000, 500}, 0, 0, 0, RED};

    Tri2D fullscreentritop =
        (Tri2D){(Vector2){0, 0}, (Vector2){1920, 0}, (Vector2){1920, 1080}, 0, 0, 0, BLUE};
    Tri2D fullscreentribottom =
        (Tri2D){(Vector2){0, 0}, (Vector2){0, 1080}, (Vector2){1920, 1080}, 0, 0, 0, RED};

    Tri2D blank = (Tri2D){(Vector2){-10, -10}, (Vector2){-10, -10}, (Vector2){-10, -10}};

    Tri2D norm = (Tri2D){(Vector2){500, 50}, (Vector2){0, 0}, (Vector2){250, 500}, 0, 0, 0, GREEN};

    bool run3d = true;

    Object3D t = ReadObjectFromFile("singletri.obj");
    if (true) {
        for (int i = 0; i < t.triangles->length; i++) {
            // printf("t: %f\n", t.triangles->arr[i].a.x);
            TriArrayAppend(&tarr, t.triangles->arr[i]);
            // printf("%s\n",tarr.arr[i].material->name);
        }
    }

    while (!WindowShouldClose() && run3d) {
        float frametime = GetFrameTime();
        CtrlLocalCam(&camera, frametime);
        LocalCamApplyVelo(&camera, frametime);
        TransformedTris.length = 0;
        Tri2Darr.length = 0;

        for (int i = 0; i < tarr.length; i++) {
            TriArrayAppend(&TransformedTris, TriTransformWithCam(&tarr.arr[i], &camera));
        }

        for (int i = 0; i < TransformedTris.length; i++) {
            if ((TransformedTris.arr[i].a.z < 0) && (TransformedTris.arr[i].b.z < 0) &&
                (TransformedTris.arr[i].c.z < 0)) {
                Tri2DArrayAppend(&Tri2Darr, ConvertTriToTri2D(&TransformedTris.arr[i]));
				//PrintTri2D(Tri2Darr.arr[Tri2Darr.length-1]);
            }
        }

        /* for (int i = 0; i < RENDERHEIGHT*RENDERWIDTH; i ++) { */
        /*     ZBuff[i] = (Zee){10000,NULL}; */
        /* } */
        memset(display, 0, sizeof(display));
        memset(ZBuff, 0, sizeof(Zee) * 1920 * 1080);
        for (int i = 0; i < RENDERHEIGHT * RENDERWIDTH; i++) {
            ZBuff[i].depth = 10000000;
        }

        BeginTextureMode(uiraylibtexture);
		ClearBackground(BLANK);
        for (int i = 0; i < Tri2Darr.length; i++) {
            DrawTriZuff(ZBuff, &Tri2Darr.arr[i]);
        }

        /* if (IsKeyDown(KEY_H)) { */
        /*     if (ZBuff[IndexOfZBuff(RENDERWIDTH / 2, RENDERHEIGHT / 2)].tri != 0) { */
        /*         // ZBuff[IndexOfZBuff(RENDERWIDTH/2, RENDERHEIGHT/2)].tri->color = RED; */
        /*         printf("%f\n", ZBuff[IndexOfZBuff(RENDERWIDTH / 2, RENDERHEIGHT / 2)].tri->a.x);
         */
        /*     } */
        /* } */

        // FillTopFlatZbuffer(ZBuff, &funners, &funners);
        // FillBottomFlatZbuffer(ZBuff,&funners2);

        // FillTopFlatZbuffer(ZBuff, &fullscreentritop);
        // FillBottomFlatZbuffer(ZBuff, &fullscreentribottom);

        // DrawTriZuff(ZBuff, &norm);
        // DrawTriZuff(ZBuff, TriTransformWithCam(&norm, &camera));

        int index = 0;
        for (int y = 0; y < RENDERHEIGHT; y++) {
            for (int x = 0; x < RENDERWIDTH; x++) {

                // REMOVE THE IF AND REPALCE THE VALUE THAT DISPLAY IS SET TO WITH DIRECT COLOR WHEN
                // DONE DEBUGGING
                // if (ZBuff[IndexOfZBuff(x, y)].tri != 0) {
                display[index] = ZBuff[index].c;
                //}
                index = index + 1;
            }
        }
        // printf("%d\n",t.mats->length);
        // printf("%d\n",t.mats->arr[t.mats->length -1].h);
        /* index = 0; */
        /* for (int y = 0; y < t.mats->arr[0].h; y++) { */
        /*     for (int x = 0; x < t.mats->arr[0].w; x++) { */
        /*         display[(y * RENDERWIDTH) + x] = t.mats->arr[0].tex[index]; */
        /*         index = index + 1; */
        /*     } */
        /* } */


        // gui stuff
        EndTextureMode();

        // Copytexture to main display :0

        BeginDrawing();
        ClearBackground(BLACK);
        UpdateTexture(render3dtexture.texture, display);

        // Copies render3dtexture to screen
        DrawTexturePro(
            render3dtexture.texture,
            (Rectangle){0, 0, render3dtexture.texture.width, render3dtexture.texture.height},
            (Rectangle){0, 0, SCREENWIDTH, SCREENHEIGHT}, (Vector2){0, 0}, 0, WHITE);

        // Copies uiraylibtexture to screen (not this is not the texture used
        // for 3d stuff
        DrawTexturePro(
            uiraylibtexture.texture,
            (Rectangle){0, 0, uiraylibtexture.texture.width, -uiraylibtexture.texture.height},
            (Rectangle){0, 0, SCREENWIDTH, SCREENHEIGHT}, (Vector2){0, 0}, 0, WHITE);

        char fpstext[40];
        sprintf(fpstext, "%d", GetFPS());
        DrawText(fpstext, 100, 0, 20, WHITE);

        EndDrawing();
    }

    CloseWindow();
    return 0;
}
